﻿using System;
using System.Reflection;
using System.Windows.Forms;

// ------------------------------------------------------------------
// Wraps System.Windows.Forms.OpenFileDialog to make it present
// a vista-style dialog.
// ------------------------------------------------------------------

namespace FolderSelect
{
    /// <summary>
    ///     Wraps System.Windows.Forms.OpenFileDialog to make it present
    ///     a vista-style dialog.
    /// </summary>
    public class FolderSelectDialog
    {
        // Wrapped dialog
        private readonly OpenFileDialog _ofd;

        /// <summary>
        ///     Default constructor
        /// </summary>
        public FolderSelectDialog()
        {
            _ofd = new OpenFileDialog
            {
                Filter = "Folders|\n",
                AddExtension = false,
                CheckFileExists = false,
                DereferenceLinks = true,
                Multiselect = false
            };
        }

        #region Properties

        /// <summary>
        ///     Gets/Sets the initial folder to be selected. A null value selects the current directory.
        /// </summary>
        public string InitialDirectory
        {
            get { return _ofd.InitialDirectory; }
            set { _ofd.InitialDirectory = string.IsNullOrEmpty(value) ? Environment.CurrentDirectory : value; }
        }

        /// <summary>
        ///     Gets/Sets the title to show in the dialog
        /// </summary>
        public string Title
        {
            get { return _ofd.Title; }
            set { _ofd.Title = value ?? "Select a folder"; }
        }

        /// <summary>
        ///     Gets the selected folder
        /// </summary>
        public string FileName
        {
            get { return _ofd.FileName; }
        }

        #endregion

        #region Methods

        /// <summary>
        ///     Shows the dialog
        /// </summary>
        /// <returns>True if the user presses OK else false</returns>
        public bool ShowDialog()
        {
            return ShowDialog(IntPtr.Zero);
        }

        /// <summary>
        ///     Shows the dialog
        /// </summary>
        /// <param name="hWndOwner">Handle of the control to be parent</param>
        /// <returns>True if the user presses OK else false</returns>
        public bool ShowDialog(IntPtr hWndOwner)
        {
            bool flag;

            if (Environment.OSVersion.Version.Major >= 6)
            {
                var r = new Reflector("System.Windows.Forms");

                uint num = 0;
                var typeIFileDialog = r.GetType("FileDialogNative.IFileDialog");
                var dialog = r.Call(_ofd, "CreateVistaDialog");
                r.Call(_ofd, "OnBeforeVistaDialog", dialog);

                var options = (uint) r.CallAs(typeof (FileDialog), _ofd, "GetOptions");
                options |= (uint) r.GetEnum("FileDialogNative.FOS", "FOS_PICKFOLDERS");
                r.CallAs(typeIFileDialog, dialog, "SetOptions", options);

                var pfde = r.New("FileDialog.VistaDialogEvents", _ofd);
                object[] parameters = {pfde, num};
                r.CallAs2(typeIFileDialog, dialog, "Advise", parameters);
                num = (uint) parameters[1];
                try
                {
                    var num2 = (int) r.CallAs(typeIFileDialog, dialog, "Show", hWndOwner);
                    flag = 0 == num2;
                }
                finally
                {
                    r.CallAs(typeIFileDialog, dialog, "Unadvise", num);
                    GC.KeepAlive(pfde);
                }
            }
            else
            {
                var fbd = new FolderBrowserDialog
                {
                    Description = Title,
                    SelectedPath = InitialDirectory,
                    ShowNewFolderButton = false
                };

                if (fbd.ShowDialog(new WindowWrapper(hWndOwner)) != DialogResult.OK) return false;
                _ofd.FileName = fbd.SelectedPath;
                flag = true;
            }

            return flag;
        }

        #endregion
    }

    /// <summary>
    ///     Creates IWin32Window around an IntPtr
    /// </summary>
    public class WindowWrapper : IWin32Window
    {
        private readonly IntPtr _hwnd;

        /// <summary>
        ///     Constructor
        /// </summary>
        /// <param name="handle">Handle to wrap</param>
        public WindowWrapper(IntPtr handle)
        {
            _hwnd = handle;
        }

        /// <summary>
        ///     Original ptr
        /// </summary>
        public IntPtr Handle
        {
            get { return _hwnd; }
        }
    }
}

/// <summary>
///     This class is from the Front-End for Dosbox and is used to present a 'vista' dialog box to select folders.
///     Being able to use a vista style dialog box to select folders is much better then using the shell folder browser.
///     http://code.google.com/p/fed/
///     Example:
///     var r = new Reflector("System.Windows.Forms");
/// </summary>
public class Reflector
{
    #region variables

    private readonly Assembly _mAsmb;
    private readonly string _mNs;

    #endregion

    #region Constructors

    /// <summary>
    ///     Constructor
    /// </summary>
    /// <param name="ns">The namespace containing types to be used</param>
    public Reflector(string ns) : this(ns, ns){}

    /// <summary>
    ///     Constructor
    /// </summary>
    /// <param name="an">A specific assembly name (used if the assembly name does not tie exactly with the namespace)</param>
    /// <param name="ns">The namespace containing types to be used</param>
    public Reflector(string an, string ns)
    {
        _mNs = ns;
        _mAsmb = null;
        foreach (var aN in Assembly.GetExecutingAssembly().GetReferencedAssemblies())
            if (aN.FullName.StartsWith(an))
            {
                _mAsmb = Assembly.Load(aN);
                break;
            }
    }

    #endregion

    #region Methods

    /// <summary>
    ///     Return a Type instance for a type 'typeName'
    /// </summary>
    /// <param name="typeName">The name of the type</param>
    /// <returns>A type instance</returns>
    public Type GetType(string typeName)
    {
        Type type = null;
        var names = typeName.Split('.');

        if (names.Length > 0)
            type = _mAsmb.GetType(_mNs + "." + names[0]);

        if(type!=null)
            for (var i = 1; i < names.Length; ++i)
            {
                type = type.GetNestedType(names[i], BindingFlags.NonPublic);
            }

        return type;
    }

    /// <summary>
    ///     Create a new object of a named type passing along any params
    /// </summary>
    /// <param name="name">The name of the type to create</param>
    /// <param name="parameters"></param>
    /// <returns>An instantiated type</returns>
    public object New(string name, params object[] parameters)
    {
        var type = GetType(name);

        var ctorInfos = type.GetConstructors();
        foreach (var ci in ctorInfos)
        {
            try
            {
                return ci.Invoke(parameters);
            }
            catch
            {
            }
        }

        return null;
    }

    /// <summary>
    ///     Calls method 'func' on object 'obj' passing parameters 'parameters'
    /// </summary>
    /// <param name="obj">The object on which to excute function 'func'</param>
    /// <param name="func">The function to execute</param>
    /// <param name="parameters">The parameters to pass to function 'func'</param>
    /// <returns>The result of the function invocation</returns>
    public object Call(object obj, string func, params object[] parameters)
    {
        return Call2(obj, func, parameters);
    }

    /// <summary>
    ///     Calls method 'func' on object 'obj' passing parameters 'parameters'
    /// </summary>
    /// <param name="obj">The object on which to excute function 'func'</param>
    /// <param name="func">The function to execute</param>
    /// <param name="parameters">The parameters to pass to function 'func'</param>
    /// <returns>The result of the function invocation</returns>
    public object Call2(object obj, string func, object[] parameters)
    {
        return CallAs2(obj.GetType(), obj, func, parameters);
    }

    /// <summary>
    ///     Calls method 'func' on object 'obj' which is of type 'type' passing parameters 'parameters'
    /// </summary>
    /// <param name="type">The type of 'obj'</param>
    /// <param name="obj">The object on which to excute function 'func'</param>
    /// <param name="func">The function to execute</param>
    /// <param name="parameters">The parameters to pass to function 'func'</param>
    /// <returns>The result of the function invocation</returns>
    public object CallAs(Type type, object obj, string func, params object[] parameters)
    {
        return CallAs2(type, obj, func, parameters);
    }

    /// <summary>
    ///     Calls method 'func' on object 'obj' which is of type 'type' passing parameters 'parameters'
    /// </summary>
    /// <param name="type">The type of 'obj'</param>
    /// <param name="obj">The object on which to excute function 'func'</param>
    /// <param name="func">The function to execute</param>
    /// <param name="parameters">The parameters to pass to function 'func'</param>
    /// <returns>The result of the function invocation</returns>
    public object CallAs2(Type type, object obj, string func, object[] parameters)
    {
        var methInfo = type.GetMethod(func, BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic);
        return methInfo.Invoke(obj, parameters);
    }

    /// <summary>
    ///     Returns the value of property 'prop' of object 'obj'
    /// </summary>
    /// <param name="obj">The object containing 'prop'</param>
    /// <param name="prop">The property name</param>
    /// <returns>The property value</returns>
    public object Get(object obj, string prop)
    {
        return GetAs(obj.GetType(), obj, prop);
    }

    /// <summary>
    ///     Returns the value of property 'prop' of object 'obj' which has type 'type'
    /// </summary>
    /// <param name="type">The type of 'obj'</param>
    /// <param name="obj">The object containing 'prop'</param>
    /// <param name="prop">The property name</param>
    /// <returns>The property value</returns>
    public object GetAs(Type type, object obj, string prop)
    {
        var propInfo = type.GetProperty(prop,
            BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic);
        return propInfo.GetValue(obj, null);
    }

    /// <summary>
    ///     Returns an enum value
    /// </summary>
    /// <param name="typeName">The name of enum type</param>
    /// <param name="name">The name of the value</param>
    /// <returns>The enum value</returns>
    public object GetEnum(string typeName, string name)
    {
        var type = GetType(typeName);
        var fieldInfo = type.GetField(name);
        return fieldInfo.GetValue(null);
    }

    #endregion
}